package geometry.part1;

import java.awt.geom.Point2D;

/**
 * 直線 a * x + b * y + c = 0 を表現するクラス
 */
public class Line {
    public double a;
    public double b;
    public double c;

    /**
     * 一般形のパラメータから直線を作成する
     * @param a xの係数
     * @param b yの係数
     * @param c 定数項
     */
    public Line(double a, double b, double c) {
        this.a = a;
        this.b = b;
        this.c = c;
    }

    /**
     * 2点(x1, y1), (x2, y2)を通る直線を作成する
     * @param x1 1点目のx座標
     * @param y1 1点目のy座標
     * @param x2 2点目のx座標
     * @param y2 2点目のy座標
     * @return 直線
     */
    public static Line fromPoints(double x1, double y1, double x2, double y2) {
        double dx = x2 - x1;
        double dy = y2 - y1;
        return new Line(dy, -dx, dx * y1 - dy * x1);
    }

    /**
     * 与えられた直線との交点を返す
     * @param l 直線
     * @return 交点。2直線が平行の場合はnull
     */
    public Point2D getIntersectionPoint(Line l) {
        double d = a * l.b - l.a * b;
        if (d == 0.0) {
            return null;
        }
        double x = (b * l.c - l.b * c) / d;
        double y = (l.a * c - a * l.c) / d;
        return new Point2D.Double(x, y);
    }

    @Override
    public String toString() {
        return "a = " + a + ", b = " + b + ", c = " + c;
    }
}
