package geometry.part10;

import geometry.common.Demo;
import geometry.common.DemoFrame;
import geometry.common.DemoPanel;
import geometry.part10.PolygonIntersectionCalculator.Edge;
import geometry.part10.PolygonIntersectionCalculator.Status;
import geometry.part10.PolygonIntersectionCalculator.StepResult;
import geometry.part2.LineSegment;
import geometry.part6.ConvexPolygon;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JToolBar;

public class PolygonIntersectionDemo {
    public static void main(String[] args) {
        Demo.launch(PolygonIntersectionFrame.class, "凸多角形の交差計算");
    }

    public static class PolygonIntersectionFrame extends DemoFrame {
        private static final long serialVersionUID = 1L;

        enum MoveTarget {
            VERTEX1, VERTEX2, WHOLE1, WHOLE2
        }

        List<Point2D> points1;
        List<Point2D> points2;
        List<Point2D> savedPoints1;
        List<Point2D> savedPoints2;
        ConvexPolygon polygon1;
        ConvexPolygon polygon2;
        Status status;
        StepResult stepResult;
        List<Edge> leftResult;
        List<Edge> rightResult;

        Point pressPoint;
        MoveTarget workingTarget;
        int vertexIndex;

        private void createInitialData() {
            points1 = new ArrayList<Point2D>();
            points2 = new ArrayList<Point2D>();

            points1.add(new Point2D.Double(290, 55));
            points1.add(new Point2D.Double(520, 170));
            points1.add(new Point2D.Double(350, 350));
            points1.add(new Point2D.Double(125, 400));
            points1.add(new Point2D.Double(100, 225));

            points2.add(new Point2D.Double(370, 130));
            points2.add(new Point2D.Double(510, 410));
            points2.add(new Point2D.Double(260, 500));
            points2.add(new Point2D.Double(200, 315));
            points2.add(new Point2D.Double(255, 190));

            polygon1 = createPolygon(points1);
            polygon2 = createPolygon(points2);
        }

        @Override
        protected void initialize() {
            createInitialData();

            JButton nextButton = new JButton("次のイベント");
            JButton resetButton = new JButton("走査線のリセット");
            nextButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    next();
                }
            });
            resetButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    reset();
                }
            });

            JToolBar toolBar = new JToolBar();
            toolBar.setFloatable(false);
            toolBar.add(nextButton);
            toolBar.add(resetButton);

            final IntersectionPanel panel = new IntersectionPanel();
            panel.addMouseListener(new MouseAdapter() {
                @Override
                public void mousePressed(MouseEvent e) {
                    pressPoint = e.getPoint();
                    savedPoints1 = new ArrayList<Point2D>(points1);
                    savedPoints2 = new ArrayList<Point2D>(points2);
                }
            });
            panel.addMouseMotionListener(new MouseAdapter() {
                @Override
                public void mouseMoved(MouseEvent e) {
                    double x = e.getX();
                    double y = e.getY();
                    vertexIndex = getHitVertexIndex(points1, x, y);
                    if (vertexIndex >= 0) {
                        workingTarget = MoveTarget.VERTEX1;
                    } else {
                        vertexIndex = getHitVertexIndex(points2, x, y);
                        if (vertexIndex >= 0) {
                            workingTarget = MoveTarget.VERTEX2;
                        } else if (hitsEdge(points1, x, y)) {
                            workingTarget = MoveTarget.WHOLE1;
                        } else if (hitsEdge(points2, x, y)) {
                            workingTarget = MoveTarget.WHOLE2;
                        } else {
                            workingTarget = null;
                        }
                    }
                    repaint();
                }

                @Override
                public void mouseDragged(MouseEvent e) {
                    if (workingTarget == null) {
                        return;
                    }
                    double dx = e.getX() - pressPoint.getX();
                    double dy = e.getY() - pressPoint.getY();
                    List<Point2D> newPoints1 = new ArrayList<Point2D>(
                            savedPoints1);
                    List<Point2D> newPoints2 = new ArrayList<Point2D>(
                            savedPoints2);
                    switch (workingTarget) {
                    case VERTEX1:
                        Point2D base = savedPoints1.get(vertexIndex);
                        newPoints1.set(vertexIndex, new Point2D.Double(base
                                .getX()
                                + dx, base.getY() + dy));
                        break;
                    case VERTEX2:
                        base = savedPoints2.get(vertexIndex);
                        newPoints2.set(vertexIndex, new Point2D.Double(base
                                .getX()
                                + dx, base.getY() + dy));
                        break;
                    case WHOLE1:
                        for (int i = 0; i < points1.size(); i++) {
                            base = savedPoints1.get(i);
                            newPoints1.set(i, new Point2D.Double(base.getX()
                                    + dx, base.getY() + dy));
                        }
                        break;
                    case WHOLE2:
                        for (int i = 0; i < points2.size(); i++) {
                            base = savedPoints2.get(i);
                            newPoints2.set(i, new Point2D.Double(base.getX()
                                    + dx, base.getY() + dy));
                        }
                        break;
                    }
                    try {
                        ConvexPolygon newPolygon1 = createPolygon(newPoints1);
                        ConvexPolygon newPolygon2 = createPolygon(newPoints2);
                        polygon1 = newPolygon1;
                        polygon2 = newPolygon2;
                        points1 = newPoints1;
                        points2 = newPoints2;
                    } catch (IllegalArgumentException ex) {
                    }
                    reset();
                }
            });

            Container c = getContentPane();
            c.setLayout(new BorderLayout());
            c.add(toolBar, BorderLayout.NORTH);
            c.add(panel, BorderLayout.CENTER);

            reset();
        }

        private ConvexPolygon createPolygon(List<Point2D> points) {
            return new ConvexPolygon(new ArrayList<Point2D>(points));
        }

        private int getHitVertexIndex(List<Point2D> points, double x, double y) {
            int size = points.size();
            for (int i = 0; i < size; i++) {
                Point2D p = points.get(i);
                if (p.distance(x, y) <= 8) {
                    return i;
                }
            }
            return -1;
        }

        private boolean hitsEdge(List<Point2D> points, double x, double y) {
            int size = points.size();
            for (int i = 0; i < size; i++) {
                Point2D p1 = points.get(i);
                Point2D p2 = points.get((i + 1) % size);
                if (Line2D.ptSegDist(p1.getX(), p1.getY(), p2.getX(),
                        p2.getY(), x, y) <= 8) {
                    return true;
                }
            }
            return false;
        }

        private void first() {
            PolygonIntersectionCalculator calculator = new PolygonIntersectionCalculator();
            stepResult = calculator.firstPass(polygon1, polygon2, status,
                    leftResult, rightResult);
            repaint();
        }

        private void next() {
            if (stepResult == null) {
                first();
                return;
            }
            if (!stepResult.mustContinue) {
                reset();
                return;
            }
            PolygonIntersectionCalculator calculator = new PolygonIntersectionCalculator();
            stepResult = calculator.secondPass(status, leftResult, rightResult);
            repaint();
        }

        private void reset() {
            status = new Status();
            stepResult = null;
            leftResult = new ArrayList<Edge>();
            rightResult = new ArrayList<Edge>();
            repaint();
        }

        private class IntersectionPanel extends DemoPanel {
            private static final long serialVersionUID = 1L;

            @Override
            protected void doPaint(Graphics2D g) {
                PolygonIntersectionCalculator calculator = new PolygonIntersectionCalculator();
                try {
                    ConvexPolygon p = calculator.execute(polygon1, polygon2);
                    if (p != null) {
                        drawPolygon(g, p, 6);
                    }
                } catch (IllegalArgumentException e) {
                }

                drawPolygon(g, polygon1, workingTarget == MoveTarget.WHOLE1 ? 3
                        : 1);
                drawPolygon(g, polygon2, workingTarget == MoveTarget.WHOLE2 ? 3
                        : 1);

                if (workingTarget == MoveTarget.VERTEX1) {
                    drawVertex(g, polygon1.getVertex(vertexIndex), 12,
                            Color.BLACK);
                } else if (workingTarget == MoveTarget.VERTEX2) {
                    drawVertex(g, polygon2.getVertex(vertexIndex), 12,
                            Color.BLACK);
                }

                drawSegments(g, leftResult, Color.RED);
                drawSegments(g, rightResult, Color.BLUE);

                if (stepResult != null) {
                    double sweepY = stepResult.nextSweepY;
                    g
                            .setStroke(new BasicStroke(1,
                                    BasicStroke.CAP_SQUARE,
                                    BasicStroke.JOIN_BEVEL, 1f, new float[] {
                                            2, 3 }, 1));
                    g.setPaint(Color.BLACK);
                    g.draw(new Line2D.Double(0, sweepY, getWidth(), sweepY));
                }
            }

            private void drawSegments(Graphics2D g, List<Edge> edges,
                    Color color) {
                g.setStroke(new BasicStroke(3));
                g.setPaint(color);
                for (Edge e : edges) {
                    LineSegment s = e.segment;
                    g.draw(new Line2D.Double(s.x1, s.y1, s.x2, s.y2));
                }
            }

            private void drawPolygon(Graphics2D g, ConvexPolygon polygon,
                    float width) {
                g.setStroke(new BasicStroke(width));
                g.setPaint(Color.BLACK);

                int size = polygon.getEdgeCount();
                for (int i = 0; i < size; i++) {
                    LineSegment e = polygon.getEdge(i);
                    g.draw(new Line2D.Double(e.x1, e.y1, e.x2, e.y2));
                }

                for (int i = 0; i < size; i++) {
                    Point2D v = polygon.getVertex(i);
                    drawVertex(g, v, 6, Color.BLACK);
                }
            }

            private void drawVertex(Graphics2D g, Point2D p, double radius,
                    Color color) {
                g.setPaint(color);
                g.fill(new Ellipse2D.Double(p.getX() - radius / 2, p.getY()
                        - radius / 2, radius, radius));
            }
        }
    }
}
